from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING

from PIL import Image
from pydantic import BaseModel, Field, PrivateAttr

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class IndexModel(BaseModel):
    """ Index model """
    index_id: str = Field(
        default=...,
        min_length=1,
        max_length=56,
        pattern="^[a-z_][a-z0-9_]{0,55}$",
        description="The ID of the index."
    )
    description: Optional[str] = None
    index_table_schema: Optional[dict[str, Any]] = Field(default={})
    index_fields: Optional[dict[str, Any]] = Field(default={})
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    index_table_name: Optional[str] = None

    def __repr__(self):
        return f"<Index('{self.index_id}', description='{self.description}')>"


class IndexUpdate(BaseModel):
    """ Index update model """
    description: Optional[str] = None
    index_table_schema: Optional[dict[str, Any]] = None
    index_fields: Optional[dict[str, Any]] = None


class Index(IndexModel):
    __doc__ = IndexModel.__doc__
    _client: Optional["LexyClient"] = PrivateAttr(default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._client = data.pop('client', None)

    @property
    def client(self) -> "LexyClient":
        if not self._client:
            raise ValueError("API client has not been set.")
        return self._client

    def query(self,
              query_text: str = None,
              query_image: Image.Image | str = None,
              query_field: str = "embedding",
              k: int = 5,
              return_fields: list[str] = None,
              return_document: bool = False,
              embedding_model: str = None) -> list[dict]:
        """ Synchronously query an index.

        Args:
            query_text (str): The query text.
            query_image (Image.Image | str): The query image. Can be a PIL Image object or a path to an image.
            query_field (str, optional): The field to query. Defaults to "embedding".
            k (int, optional): The number of records to return. Defaults to 5.
            return_fields (list[str], optional): The fields to return. Defaults to None, which returns all fields. To
                return fields from the linked document, use "document.<field_name>".
            return_document (bool, optional): Whether to return the document object. Defaults to False.
            embedding_model (str, optional): The name of the embedding model to use. Defaults to None, which uses the
                embedding model associated with `index_id.query_field`.

        Returns:
            Results: A list of query results.
        """
        return self.client.index.query_index(query_text=query_text,
                                             query_image=query_image,
                                             index_id=self.index_id,
                                             query_field=query_field,
                                             k=k,
                                             return_fields=return_fields,
                                             return_document=return_document,
                                             embedding_model=embedding_model)

    def list_records(self, document_id: Optional[str] = None) -> list[dict]:
        """ Synchronously list all records in the index.

        Args:
            document_id (str, optional): The document ID to filter by. Defaults to None.

        Returns:
            list[dict]: A list of records.
        """
        return self.client.index.list_index_records(self.index_id, document_id=document_id)



""" Client for interacting with the Transformer API. """

from typing import Optional, TYPE_CHECKING

import httpx

from lexy_py.exceptions import handle_response
from .models import Transformer, TransformerUpdate
from lexy_py.document.models import Document

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class TransformerClient:
    """
    This class is used to interact with the Lexy Transformer API.

    Attributes:
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
    """

    def __init__(self, lexy_client: "LexyClient") -> None:
        self._lexy_client = lexy_client

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._lexy_client.aclient

    @property
    def client(self) -> httpx.Client:
        return self._lexy_client.client

    def list_transformers(self) -> list[Transformer]:
        """ Synchronously get a list of all transformers.

        Returns:
            list[Transformer]: A list of all transformers.
        """
        r = self.client.get("/transformers")
        handle_response(r)
        return [Transformer(**transformer, client=self._lexy_client) for transformer in r.json()]

    async def alist_transformers(self) -> list[Transformer]:
        """ Asynchronously get a list of all transformers.

        Returns:
            list[Transformer]: A list of all transformers.
        """
        r = await self.aclient.get("/transformers")
        handle_response(r)
        return [Transformer(**transformer, client=self._lexy_client) for transformer in r.json()]

    def get_transformer(self, transformer_id: str) -> Transformer:
        """ Synchronously get a transformer.

        Args:
            transformer_id (str): The ID of the transformer to get.

        Returns:
            Transformer: The transformer.
        """
        r = self.client.get(f"/transformers/{transformer_id}")
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    async def aget_transformer(self, transformer_id: str) -> Transformer:
        """ Asynchronously get a transformer.

        Args:
            transformer_id (str): The ID of the transformer to get.

        Returns:
            Transformer: The transformer.
        """
        r = await self.aclient.get(f"/transformers/{transformer_id}")
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    def add_transformer(self,
                        transformer_id: str,
                        path: str,
                        description: Optional[str] = None) -> Transformer:
        """ Synchronously add a transformer.

        Args:
            transformer_id (str): The ID of the transformer to add.
            path (str): The path of the transformer to add.
            description (str, optional): The description of the transformer to add.

        Returns:
            Transformer: The added transformer.
        """
        data = {"transformer_id": transformer_id, "path": path}
        if description:
            data["description"] = description
        r = self.client.post("/transformers", json=data)
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    async def aadd_transformer(self,
                               transformer_id: str,
                               path: str,
                               description: Optional[str] = None) -> Transformer:
        """ Asynchronously add a transformer.

        Args:
            transformer_id (str): The ID of the transformer to add.
            path (str): The path of the transformer to add.
            description (str, optional): The description of the transformer to add.

        Returns:
            Transformer: The added transformer.
        """
        data = {"transformer_id": transformer_id, "path": path}
        if description:
            data["description"] = description
        r = await self.aclient.post("/transformers", json=data)
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    def update_transformer(self,
                           transformer_id: str,
                           path: Optional[str] = None,
                           description: Optional[str] = None) -> Transformer:
        """ Synchronously update a transformer.

        Args:
            transformer_id (str): The ID of the transformer to update.
            path (str, optional): The updated path of the transformer.
            description (str, optional): The updated description of the transformer.

        Returns:
            Transformer: The updated transformer.
        """
        transformer = TransformerUpdate(path=path, description=description)
        r = self.client.patch(f"/transformers/{transformer_id}", json=transformer.model_dump(exclude_none=True))
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    async def aupdate_transformer(self,
                                  transformer_id: str,
                                  path: Optional[str] = None,
                                  description: Optional[str] = None) -> Transformer:
        """ Asynchronously update a transformer.

        Args:
            transformer_id (str): The ID of the transformer to update.
            path (str, optional): The updated path of the transformer.
            description (str, optional): The updated description of the transformer.

        Returns:
            Transformer: The updated transformer.
        """
        transformer = TransformerUpdate(path=path, description=description)
        r = await self.aclient.patch(f"/transformers/{transformer_id}", json=transformer.model_dump(exclude_none=True))
        handle_response(r)
        return Transformer(**r.json(), client=self._lexy_client)

    def delete_transformer(self, transformer_id: str) -> dict:
        """ Synchronously delete a transformer.

        Args:
            transformer_id (str): The ID of the transformer to delete.
        """
        r = self.client.delete(f"/transformers/{transformer_id}")
        handle_response(r)
        return r.json()

    async def adelete_transformer(self, transformer_id: str) -> dict:
        """ Asynchronously delete a transformer.

        Args:
            transformer_id (str): The ID of the transformer to delete.
        """
        r = await self.aclient.delete(f"/transformers/{transformer_id}")
        handle_response(r)
        return r.json()

    def transform_document(self,
                           transformer_id: str,
                           document: Document | dict,
                           transformer_params: dict = None,
                           content_only: bool = False) -> dict:
        """ Synchronously transform a document.

        Args:
            transformer_id (str): The ID of the transformer to use.
            document (Document | dict): The document to transform.
            transformer_params (dict, optional): The transformer parameters. Defaults to None.
            content_only (bool, optional): Whether to submit only the document content and not the document itself.
                Use this option when the transformer doesn't accept Document objects as inputs. Defaults to False.

        Returns:
            dict: A dictionary containing the generated task ID and the result of the transformer.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.transformer.transform_document("text.counter.word_counter", {"content": "Hello world!"})
            {'task_id': '65ecd2f7-bac4-4747-9e65-a6d21a72f585', 'result': [2, 'world!']}
        """
        if isinstance(document, dict):
            document = Document(**document)
        data = {"document": document.model_dump(mode='json')}
        if transformer_params:
            data["transformer_params"] = transformer_params
        if content_only:
            data["content_only"] = content_only
        r = self.client.post(f"/transformers/{transformer_id}", json=data)
        handle_response(r)
        return r.json()



from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING

from PIL import Image
from pydantic import BaseModel, Field, PrivateAttr

from lexy_py.document.models import Document

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class CollectionModel(BaseModel):
    """ Collection model """
    collection_name: str = Field(
        title="Collection Name",
        description="The name of the collection. Must be unique across collections.",
        min_length=1,
        max_length=56,
        pattern="^[a-z_][a-z0-9_]{0,55}$",
    )
    description: Optional[str] = None
    config: Optional[dict[str, Any]] = Field(default={})
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    collection_id: Optional[str] = None

    def __repr__(self):
        return f"<Collection('{self.collection_name}', id='{self.collection_id}', description='{self.description}')>"


class CollectionUpdate(BaseModel):
    """ Collection update model """
    collection_name: Optional[str] = Field(
        default=None,
        title="Collection Name",
        description="The name of the collection. Must be unique across collections.",
        min_length=1,
        max_length=56,
        pattern="^[a-z_][a-z0-9_]{0,55}$",
    )
    description: Optional[str] = None
    config: Optional[dict[str, Any]] = None


class Collection(CollectionModel):
    __doc__ = CollectionModel.__doc__
    _client: Optional["LexyClient"] = PrivateAttr(default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._client = data.pop('client', None)

    @property
    def client(self) -> "LexyClient":
        if not self._client:
            raise ValueError("API client has not been set.")
        return self._client

    def add_documents(self,
                      docs: Document | dict | list[Document | dict],
                      *,
                      batch_size: int = 100) -> list[Document]:
        """ Synchronously add documents to the collection in batches.

        Args:
            docs (Document | dict | list[Document | dict]): The documents to add.
            batch_size (int): The number of documents to add in each batch. Defaults to 100.

        Returns:
            Documents: A list of created documents.
        """
        return self.client.document.add_documents(docs, collection_id=self.collection_id, batch_size=batch_size)

    # TODO: add pagination
    def list_documents(self, *, limit: int = 100, offset: int = 0) -> list[Document]:
        """ Synchronously get a list of documents in the collection.

        Args:
            limit (int): The maximum number of documents to return. Defaults to 100. Maximum allowed is 1000.
            offset (int): The offset to start from. Defaults to 0.

        Returns:
            Documents: A list of documents in the collection.
        """
        return self.client.document.list_documents(collection_id=self.collection_id, limit=limit, offset=offset)

    def upload_documents(self,
                         files: str | Image.Image | list[str | Image.Image],
                         filenames: str | list[str] = None,
                         *,
                         batch_size: int = 5) -> list[Document]:
        """ Synchronously upload files to the collection in batches.

        Args:
            files (str | Image.Image | list[str | Image.Image]): The files to upload. Can be a single instance or a
                list of a string containing the path to a file or an `Image.Image` object.
            filenames (str | list[str], optional): The filenames of the files to upload. Defaults to None.
            batch_size (int): The number of files to upload in each batch. Defaults to 5.

        Returns:
            Documents: A list of created documents.

        Raises:
            TypeError: If an input file type is invalid.
            ValueError: If the length of the filenames list does not match the length of the files list.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> collection = lx.create_collection(collection_name='my_file_collection')
            >>> collection.upload_documents(
            ...     files=[
            ...         'lexy/sample_data/images/lexy-dalle.jpeg',
            ...         'lexy/sample_data/images/lexy.png',
            ...         'dais2023-233180.pdf',
            ...         'gwdemo30.mp4',
            ...         'kindle2.html',
            ...     ]
            ... )
        """
        return self.client.document.upload_documents(files=files,
                                                     filenames=filenames,
                                                     collection_id=self.collection_id,
                                                     batch_size=batch_size)



from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING

from pydantic import BaseModel, Field, PrivateAttr

from lexy_py.document.models import Document

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class TransformerModel(BaseModel):
    """ Transformer model """
    transformer_id: str = Field(..., min_length=1, max_length=255, pattern=r"^[a-zA-Z][a-zA-Z0-9_.-]+$")
    path: Optional[str] = Field(default=None, min_length=1, max_length=255, pattern=r"^[a-zA-Z][a-zA-Z0-9_.]+$")
    description: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    def __repr__(self):
        return f"<Transformer('{self.transformer_id}', description='{self.description}')>"


class TransformerUpdate(BaseModel):
    """ Transformer update model """
    path: Optional[str] = None
    description: Optional[str] = None


class Transformer(TransformerModel):
    __doc__ = TransformerModel.__doc__
    _client: Optional["LexyClient"] = PrivateAttr(default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._client = data.pop('client', None)

    @property
    def client(self) -> "LexyClient":
        if not self._client:
            raise ValueError("API client has not been set.")
        return self._client

    def transform_document(self,
                           document: Document | dict,
                           transformer_params: dict = None,
                           content_only: bool = False) -> dict:
        """ Synchronously transform a document.

        Args:
            document (Document | dict): The document to transform.
            transformer_params (dict, optional): The transformer parameters. Defaults to None.
            content_only (bool, optional): Whether to submit only the document content and not the document itself.
                Use this option when the transformer doesn't accept Document objects as inputs. Defaults to False.

        Returns:
            dict: A dictionary containing the generated task ID and the result of the transformer.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> minilm = lx.get_transformer('text.embeddings.minilm')
            >>> minilm.transform_document({'content': 'Good morning!'})
            {'task_id': '449d9d79-4a57-4191-95d3-9c38955c8ced',
             'result': [-0.03085244633257389, 0.028894789516925812, ...]}
        """
        return self.client.transformer.transform_document(transformer_id=self.transformer_id,
                                                          document=document,
                                                          transformer_params=transformer_params,
                                                          content_only=content_only)



import base64
import textwrap
from datetime import datetime
from io import BytesIO
from typing import Any, Optional, TYPE_CHECKING

from PIL import Image
from pydantic import BaseModel, Field, PrivateAttr

from lexy_py.storage import presigned_url_is_expired

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class DocumentModel(BaseModel):
    """ Document model """
    document_id: Optional[str] = None
    content: str = Field(...)
    meta: Optional[dict[Any, Any]] = Field(default={})
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    collection_id: Optional[str] = None

    def __repr__(self):
        return f'<Document("{textwrap.shorten(self.content, 100, placeholder="...")}")>'

    def __init__(self, content: str, **data: Any):
        super().__init__(content=content, **data)


class DocumentCreate(BaseModel):
    """ Document create model """
    content: str = Field(...)
    meta: Optional[dict[Any, Any]] = Field(default={})


class DocumentUpdate(BaseModel):
    """ Document update model """
    content: Optional[str] = None
    meta: Optional[dict[Any, Any]] = None


class Document(DocumentModel):
    __doc__ = DocumentModel.__doc__
    _client: Optional["LexyClient"] = PrivateAttr(default=None)
    _image: Optional["Image"] = PrivateAttr(default=None)
    _urls: Optional[dict] = PrivateAttr(default=None)

    def __init__(self, content: str, **data: Any):
        super().__init__(content=content, **data)
        self._client = data.pop('client', None)

    @property
    def client(self) -> "LexyClient":
        if not self._client:
            raise ValueError("API client has not been set.")
        return self._client

    @property
    def image(self) -> Image:
        if not self._image:
            self._image = self.meta.get('image', {}).get('im')
        if not self._image:
            base64_str = self.meta.get('image', {}).get('base64')
            if base64_str:
                self._image = self.image_from_base64_str(base64_str)
        if not self._image:
            if self.object_url:
                self._image = self.image_from_url(self.object_url)
        return self._image

    @property
    def object_url(self) -> str | None:
        if self._urls is None:
            self._refresh_urls()
        url = self._urls.get('object', None)
        # check if url is expired and refresh if needed
        storage_service = self.meta.get('storage_service')
        if url and presigned_url_is_expired(url, storage_service=storage_service):
            self._refresh_urls()
            url = self._urls.get('object', None)
        return url

    def get_thumbnail_url(self, size: tuple[int, int] = (200, 200), refresh: bool = False) -> str | None:
        if self._urls is None or refresh:
            self._refresh_urls()
        url = self._urls.get('thumbnails', {}).get(f'{size[0]}x{size[1]}')
        # if that size doesn't exist, get any size
        if not url:
            url = next(iter(self._urls.get('thumbnails', {}).values()), None)
        # check if url is expired and refresh if needed
        storage_service = (self.meta.get('image', {}).get('thumbnails', {}).get(f'{size[0]}x{size[1]}', {})
                           .get('storage_service'))
        if url and presigned_url_is_expired(url, storage_service=storage_service):
            return self.get_thumbnail_url(size, refresh=True)
        return url
    thumbnail_url = property(get_thumbnail_url)

    def _refresh_urls(self):
        self._urls = self.client.document.get_document_urls(self.document_id)

    # TODO: move to future ImageDocument class
    def image_from_url(self, url: str) -> Image:
        if self._client:
            r = self.client.get(url)
        else:
            import httpx
            r = httpx.get(url)
        return Image.open(BytesIO(r.content))

    # TODO: move to future ImageDocument class
    @staticmethod
    def image_from_base64_str(base64_str: str) -> Image:
        img_bytes = base64.b64decode(base64_str)
        return Image.open(BytesIO(img_bytes))



""" Client for interacting with the Lexy API. """

import httpx

from .settings import DEFAULT_BASE_URL
from .exceptions import LexyClientError
from .binding.client import BindingClient
from .collection.client import CollectionClient
from .document.client import DocumentClient
from .index.client import IndexClient
from .transformer.client import TransformerClient


API_TIMEOUT = 10


class LexyClient:
    """
    LexyClient class

    Attributes:
        base_url (str): Base URL for the Lexy API.
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
        binding (BindingClient): Client for interacting with the Bindings API.
        collection (CollectionClient): Client for interacting with the Collections API.
        document (DocumentClient): Client for interacting with the Documents API.
        index (IndexClient): Client for interacting with the Indexes API.
        transformer (TransformerClient): Client for interacting with the Transformers API.
    """

    base_url: str
    aclient: httpx.AsyncClient
    client: httpx.Client

    binding: BindingClient
    collection: CollectionClient
    document: DocumentClient
    index: IndexClient
    transformer: TransformerClient

    def __init__(self,
                 base_url: str = DEFAULT_BASE_URL,
                 api_timeout: int = API_TIMEOUT,
                 client_kwargs: dict = None,
                 aclient_kwargs: dict = None) -> None:
        """ Initialize a LexyClient instance.

        Args:
            base_url (str, optional): Base URL for the Lexy API. Defaults to DEFAULT_BASE_URL.
            api_timeout (int, optional): Timeout in seconds for API requests. Defaults to API_TIMEOUT.
            client_kwargs (dict, optional): Keyword arguments for the synchronous API client.
            aclient_kwargs (dict, optional): Keyword arguments for the asynchronous API client.
        """
        self.base_url = base_url
        self.api_timeout = api_timeout

        client_kwargs = client_kwargs or {}
        self._client = httpx.Client(base_url=self.base_url, timeout=self.api_timeout, **client_kwargs)
        aclient_kwargs = aclient_kwargs or {}
        self._aclient = httpx.AsyncClient(base_url=self.base_url, timeout=self.api_timeout, **aclient_kwargs)

        # binding
        self.binding = BindingClient(self)
        self.create_binding = self.binding.add_binding
        self.delete_binding = self.binding.delete_binding
        self.get_binding = self.binding.get_binding
        self.list_bindings = self.binding.list_bindings
        self.update_binding = self.binding.update_binding

        # collection
        self.collection = CollectionClient(self)
        self.create_collection = self.collection.add_collection
        self.delete_collection = self.collection.delete_collection
        self.get_collection = self.collection.get_collection
        self.list_collections = self.collection.list_collections
        self.update_collection = self.collection.update_collection

        # document
        self.document = DocumentClient(self)
        self.add_documents = self.document.add_documents
        self.bulk_delete_documents = self.document.bulk_delete_documents
        self.delete_document = self.document.delete_document
        self.get_document = self.document.get_document
        self.list_documents = self.document.list_documents
        self.update_document = self.document.update_document
        self.upload_documents = self.document.upload_documents

        # index
        self.index = IndexClient(self)
        self.create_index = self.index.add_index
        self.delete_index = self.index.delete_index
        self.get_index = self.index.get_index
        self.list_indexes = self.index.list_indexes
        self.query_index = self.index.query_index
        self.update_index = self.index.update_index

        # transformer
        self.transformer = TransformerClient(self)
        self.create_transformer = self.transformer.add_transformer
        self.delete_transformer = self.transformer.delete_transformer
        self.get_transformer = self.transformer.get_transformer
        self.list_transformers = self.transformer.list_transformers
        self.update_transformer = self.transformer.update_transformer
        self.transform_document = self.transformer.transform_document

    @property
    def client(self) -> httpx.Client:
        if self._client is None:
            raise LexyClientError("Client is not set.")
        return self._client

    @client.setter
    def client(self, value) -> None:
        self._client = value

    @property
    def aclient(self) -> httpx.AsyncClient:
        if self._aclient is None:
            raise LexyClientError("AsyncClient is not set.")
        return self._aclient

    @aclient.setter
    def aclient(self, value) -> None:
        self._aclient = value

    async def __aenter__(self) -> "LexyClient":
        """ Async context manager entry point. """
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        """ Async context manager exit point. """
        await self.aclient.aclose()

    def __enter__(self) -> "LexyClient":
        """ Synchronous context manager entry point. """
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        """ Synchronous context manager exit point. """
        self.client.close()

    def get(self, url: str, **kwargs) -> httpx.Response:
        """ Synchronous GET request. """
        return self.client.get(url, **kwargs)

    async def aget(self, url: str, **kwargs) -> httpx.Response:
        """ Async GET request. """
        return await self.aclient.get(url, **kwargs)

    def post(self, url: str, **kwargs) -> httpx.Response:
        """ Synchronous POST request. """
        return self.client.post(url, **kwargs)

    async def apost(self, url: str, **kwargs) -> httpx.Response:
        """ Async POST request. """
        return await self.aclient.post(url, **kwargs)

    def patch(self, url: str, **kwargs) -> httpx.Response:
        """ Synchronous PATCH request. """
        return self.client.patch(url, **kwargs)

    async def apatch(self, url: str, **kwargs) -> httpx.Response:
        """ Async PATCH request. """
        return await self.aclient.patch(url, **kwargs)

    def delete(self, url: str, **kwargs) -> httpx.Response:
        """ Synchronous DELETE request. """
        return self.client.delete(url, **kwargs)

    async def adelete(self, url: str, **kwargs) -> httpx.Response:
        """ Async DELETE request. """
        return await self.aclient.delete(url, **kwargs)

    @property
    def bindings(self):
        """ Get a list of all bindings. """
        return self.binding.list_bindings()

    @property
    def collections(self):
        """ Get a list of all collections. """
        return self.collection.list_collections()

    @property
    def indexes(self):
        """ Get a list of all indexes. """
        return self.index.list_indexes()

    @property
    def transformers(self):
        """ Get a list of all transformers. """
        return self.transformer.list_transformers()

    def info(self):
        """ Print info about the Lexy server. """
        collections_str = "\n".join([f"\t- {c.__repr__()}" for c in self.collections])
        indexes_str = "\n".join([f"\t- {i.__repr__()}" for i in self.indexes])
        transformers_str = "\n".join([f"\t- {t.__repr__()}" for t in self.transformers])
        bindings_str = "\n".join([f"\t- {b.__repr__()}" for b in self.bindings])
        info_str = \
            f"Lexy server <{self.base_url}>\n\n" \
            f"{len(self.collections)} Collections\n{collections_str}\n" \
            f"{len(self.indexes)} Indexes\n{indexes_str}\n" \
            f"{len(self.transformers)} Transformers\n{transformers_str}\n" \
            f"{len(self.bindings)} Bindings\n{bindings_str}\n"
        print(info_str)



""" Client for interacting with the Document API. """

import mimetypes
import os
from io import BytesIO
from typing import Optional, TYPE_CHECKING

import httpx
from PIL import Image

from lexy_py.exceptions import handle_response
from .models import Document, DocumentUpdate

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class DocumentClient:
    """
    This class is used to interact with the Lexy Document API.

    Attributes:
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
    """

    def __init__(self, lexy_client: "LexyClient") -> None:
        self._lexy_client = lexy_client

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._lexy_client.aclient

    @property
    def client(self) -> httpx.Client:
        return self._lexy_client.client

    def list_documents(self,
                       *,
                       collection_name: str | None = "default",
                       collection_id: str = None,
                       limit: int = 100,
                       offset: int = 0) -> list[Document]:
        """ Synchronously get a list of documents in a collection.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to get documents from. Defaults to "default".
            collection_id (str): The ID of the collection to get documents from. Defaults to None. If provided,
                `collection_name` will be ignored.
            limit (int): The maximum number of documents to return. Defaults to 100. Maximum allowed is 1000.
            offset (int): The offset to start from. Defaults to 0.

        Returns:
            Documents: A list of documents in a collection.
        """
        if collection_id is not None:
            r = self.client.get(f"/collections/{collection_id}/documents",
                                params={"limit": limit, "offset": offset})
        else:
            r = self.client.get("/documents",
                                params={
                                    "collection_name": collection_name,
                                    "limit": limit,
                                    "offset": offset
                                })
        handle_response(r)
        return [Document(**document, client=self._lexy_client) for document in r.json()]

    async def alist_documents(self,
                              *,
                              collection_name: str = "default",
                              collection_id: str = None,
                              limit: int = 100,
                              offset: int = 0) -> list[Document]:
        """ Asynchronously get a list of documents in a collection.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to get documents from. Defaults to "default".
            collection_id (str): The ID of the collection to get documents from. Defaults to None. If provided,
                `collection_name` will be ignored.
            limit (int): The maximum number of documents to return. Defaults to 100. Maximum allowed is 1000.
            offset (int): The offset to start from. Defaults to 0.

        Returns:
            Documents: A list of documents in a collection.
        """
        if collection_id is not None:
            r = await self.aclient.get(f"/collections/{collection_id}/documents",
                                       params={"limit": limit, "offset": offset})
        else:
            r = await self.aclient.get("/documents",
                                       params={
                                           "collection_name": collection_name,
                                           "limit": limit,
                                           "offset": offset
                                       })
        handle_response(r)
        return [Document(**document, client=self._lexy_client) for document in r.json()]

    def add_documents(self,
                      docs: Document | dict | list[Document | dict],
                      *,
                      collection_name: str = "default",
                      collection_id: str = None,
                      batch_size: int = 100) -> list[Document]:
        """ Synchronously add documents to a collection in batches.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            docs (Document | dict | list[Document | dict]): The documents to add.
            collection_name (str): The name of the collection to add the documents to. Defaults to "default".
            collection_id (str): The ID of the collection to add the documents to. Defaults to None. If provided,
                `collection_name` will be ignored.
            batch_size (int): The number of documents to add in each batch. Defaults to 100.

        Returns:
            Documents: A list of created documents.

        Examples:
            Add documents to the default collection:

            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> docs_added = lx.add_documents(docs=[
            ...     {"content": "Default document 1"},
            ...     {"content": "Default document 2"}
            ... ])

            Add documents to a specific collection by name:

            >>> my_new_collection = lx.create_collection("my_new_collection")
            >>> docs_added = lx.add_documents(docs=[
            ...     {"content": "My first document"},
            ...     {"content": "My second document"}
            ... ], collection_name="my_new_collection")

            Add documents to a specific collection by ID:

            >>> docs_added = lx.add_documents(docs=[
            ...     {"content": "My third document"},
            ...     {"content": "My fourth document"}
            ... ], collection_id=my_new_collection.collection_id)

        """
        created_docs = []
        processed_docs = self._process_docs(docs)

        for i in range(0, len(processed_docs), batch_size):
            batch_docs = processed_docs[i:i + batch_size]

            if collection_id is not None:
                r = self.client.post(f"/collections/{collection_id}/documents", json=batch_docs)
            else:
                r = self.client.post("/documents", json=batch_docs, params={"collection_name": collection_name})

            handle_response(r)
            created_docs.extend(
                [Document(**document['document'], client=self._lexy_client) for document in r.json()]
            )
        return created_docs

    async def aadd_documents(self,
                             docs: Document | dict | list[Document | dict],
                             *,
                             collection_name: str = "default",
                             collection_id: str = None,
                             batch_size: int = 100) -> list[Document]:
        """ Asynchronously add documents to a collection in batches.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            docs (Document | dict | list[Document | dict]): The documents to add.
            collection_name (str): The name of the collection to add the documents to. Defaults to "default".
            collection_id (str): The ID of the collection to add the documents to. Defaults to None. If provided,
                `collection_name` will be ignored.
            batch_size (int): The number of documents to add in each batch. Defaults to 100.

        Returns:
            Documents: A list of created documents.
        """
        created_docs = []
        processed_docs = self._process_docs(docs)

        for i in range(0, len(processed_docs), batch_size):
            batch_docs = processed_docs[i:i + batch_size]

            if collection_id is not None:
                r = await self.aclient.post(f"/collections/{collection_id}/documents", json=batch_docs)
            else:
                r = await self.aclient.post("/documents", json=batch_docs,
                                            params={"collection_name": collection_name})

            handle_response(r)
            created_docs.extend(
                [Document(**document['document'], client=self._lexy_client) for document in r.json()]
            )
        return created_docs

    def add_document(self,
                     doc: Document | dict,
                     *,
                     collection_name: str = "default",
                     collection_id: str = None) -> Document:
        """ Synchronously add a document to a collection.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            doc (Document | dict): The document to add.
            collection_name (str): The name of the collection to add the document to. Defaults to "default".
            collection_id (str): The ID of the collection to add the document to. If provided, `collection_name` will
                be ignored.

        Returns:
            Document: The created document.
        """
        processed_docs = self._process_docs(doc)

        if collection_id is not None:
            r = self.client.post(f"/collections/{collection_id}/documents", json=processed_docs)
        else:
            r = self.client.post("/documents", json=processed_docs, params={"collection_name": collection_name})

        handle_response(r)
        return Document(**r.json()[0]['document'], client=self._lexy_client)

    async def aadd_document(self,
                            doc: Document | dict,
                            *,
                            collection_name: str = "default",
                            collection_id: str = None) -> Document:
        """ Asynchronously add a document to a collection.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            doc (Document | dict): The document to add.
            collection_name (str): The name of the collection to add the document to. Defaults to "default".
            collection_id (str): The ID of the collection to add the document to. If provided, `collection_name` will
                be ignored.

        Returns:
            Document: The created document.
        """
        processed_docs = self._process_docs(doc)

        if collection_id is not None:
            r = await self.aclient.post(f"/collections/{collection_id}/documents", json=processed_docs)
        else:
            r = await self.aclient.post("/documents", json=processed_docs,
                                        params={"collection_name": collection_name})

        handle_response(r)
        return Document(**r.json()[0]['document'], client=self._lexy_client)

    def get_document(self, document_id: str) -> Document:
        """ Synchronously get a document.

        Args:
            document_id (str): The ID of the document to get.

        Returns:
            Document: The document.
        """
        r = self.client.get(f"/documents/{document_id}")
        handle_response(r)
        return Document(**r.json(), client=self._lexy_client)

    async def aget_document(self, document_id: str) -> Document:
        """ Asynchronously get a document.

        Args:
            document_id (str): The ID of the document to get.

        Returns:
            Document: The document.
        """
        r = await self.aclient.get(f"/documents/{document_id}")
        handle_response(r)
        return Document(**r.json(), client=self._lexy_client)

    def update_document(self,
                        document_id: str,
                        *,
                        content: Optional[str] = None,
                        meta: Optional[dict] = None) -> Document:
        """ Synchronously update a document.

        Args:
            document_id (str): The ID of the document to update.
            content (str, optional): The new content of the document.
            meta (dict, optional): The new metadata for the document.

        Returns:
            Document: The updated document.
        """
        document = DocumentUpdate(content=content, meta=meta)
        r = self.client.patch(f"/documents/{document_id}", json=document.model_dump(exclude_none=True))
        handle_response(r)
        return Document(**r.json()['document'], client=self._lexy_client)

    async def aupdate_document(self,
                               document_id: str,
                               *,
                               content: Optional[str] = None,
                               meta: Optional[dict] = None) -> Document:
        """ Asynchronously update a document.

        Args:
            document_id (str): The ID of the document to update.
            content (str, optional): The new content of the document.
            meta (dict, optional): The new metadata for the document.

        Returns:
            Document: The updated document.
        """
        document = DocumentUpdate(content=content, meta=meta)
        r = await self.aclient.patch(f"/documents/{document_id}", json=document.model_dump(exclude_none=True))
        handle_response(r)
        return Document(**r.json()['document'], client=self._lexy_client)

    def delete_document(self, document_id: str) -> dict:
        """ Synchronously delete a document.

        Args:
            document_id (str): The ID of the document to delete.
        """
        r = self.client.delete(f"/documents/{document_id}")
        handle_response(r)
        return r.json()

    async def adelete_document(self, document_id: str) -> dict:
        """ Asynchronously delete a document.

        Args:
            document_id (str): The ID of the document to delete.
        """
        r = await self.aclient.delete(f"/documents/{document_id}")
        handle_response(r)
        return r.json()

    def bulk_delete_documents(self,
                              *,
                              collection_name: str = None,
                              collection_id: str = None) -> dict:
        """ Synchronously delete all documents from a collection.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to delete documents from.
            collection_id (str): The ID of the collection to delete documents from. If provided, `collection_name`
                will be ignored.

        Raises:
            ValueError: If neither `collection_name` nor `collection_id` are provided.
        """
        if collection_id:
            r = self.client.delete(f"/collections/{collection_id}/documents")
        elif collection_name:
            r = self.client.delete("/documents", params={"collection_name": collection_name})
        else:
            raise ValueError("Either 'collection_name' or 'collection_id' must be provided")
        handle_response(r)
        return r.json()

    def upload_documents(self,
                         files: str | Image.Image | list[str | Image.Image],
                         filenames: str | list[str] = None,
                         *,
                         collection_name: str = "default",
                         collection_id: str = None,
                         batch_size: int = 5) -> list[Document]:
        """ Synchronously upload files to a collection in batches.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            files (str | Image.Image | list[str | Image.Image]): The files to upload. Can be a single instance or a
                list of a string containing the path to a file or an `Image.Image` object.
            filenames (str | list[str], optional): The filenames of the files to upload. Defaults to None.
            collection_name (str): The name of the collection to upload the files to. Defaults to "default".
            collection_id (str): The ID of the collection to upload the files to. Defaults to None. If provided,
                `collection_name` will be ignored.
            batch_size (int): The number of files to upload in each batch. Defaults to 5.

        Returns:
            Documents: A list of created documents.

        Raises:
            TypeError: If an input file type is invalid.
            ValueError: If the length of the filenames list does not match the length of the files list.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.document.upload_documents(
            ...     files=[
            ...         'lexy/sample_data/images/lexy-dalle.jpeg',
            ...         'lexy/sample_data/images/lexy.png',
            ...         'dais2023-233180.pdf',
            ...         'gwdemo30.mp4',
            ...         'kindle2.html',
            ...     ],
            ...     collection_name='my_file_collection',
            ... )
        """
        created_docs = []
        files, filenames = self._align_filenames(files, filenames)

        # process and upload files in batches
        for i in range(0, len(files), batch_size):
            batch_files = files[i:i + batch_size]
            batch_filenames = filenames[i:i + batch_size]
            processed_files = self._process_files(batch_files, filenames=batch_filenames)

            if collection_id is not None:
                r = self.client.post(f"/collections/{collection_id}/documents/upload",
                                     files=processed_files)
            else:
                r = self.client.post("/documents/upload",
                                     files=processed_files,
                                     params={"collection_name": collection_name})

            handle_response(r)
            created_docs.extend(
                [Document(**document['document'], client=self._lexy_client) for document in r.json()]
            )
        return created_docs

    async def aupload_documents(self,
                                files: str | Image.Image | list[str | Image.Image],
                                filenames: str | list[str] = None,
                                *,
                                collection_name: str = "default",
                                collection_id: str = None,
                                batch_size: int = 5) -> list[Document]:
        """ Asynchronously upload files to a collection in batches.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            files (str | Image.Image | list[str | Image.Image]): The files to upload. Can be a single instance or a
                list of a string containing the path to a file or an `Image.Image` object.
            filenames (str | list[str], optional): The filenames of the files to upload. Defaults to None.
            collection_name (str): The name of the collection to upload the files to. Defaults to "default".
            collection_id (str): The ID of the collection to upload the files to. Defaults to None. If provided,
                `collection_name` will be ignored.
            batch_size (int): The number of files to upload in each batch. Defaults to 5.

        Returns:
            Documents: A list of created documents.

        Raises:
            TypeError: If an input file type is invalid.
            ValueError: If the length of the filenames list does not match the length of the files list.
        """
        created_docs = []
        files, filenames = self._align_filenames(files, filenames)

        # process and upload files in batches
        for i in range(0, len(files), batch_size):
            batch_files = files[i:i + batch_size]
            batch_filenames = filenames[i:i + batch_size]
            processed_files = self._process_files(batch_files, filenames=batch_filenames)

            if collection_id is not None:
                r = await self.aclient.post(f"/collections/{collection_id}/documents/upload",
                                            files=processed_files)
            else:
                r = await self.aclient.post("/documents/upload",
                                            files=processed_files,
                                            params={"collection_name": collection_name})

            handle_response(r)
            created_docs.extend(
                [Document(**document['document'], client=self._lexy_client) for document in r.json()]
            )
        return created_docs

    def get_document_urls(self, document_id: str, *, expiration: int = 3600) -> dict:
        """ Synchronously get presigned URLs for a document.

        Args:
            document_id (str): The ID of the document to get presigned URL for.
            expiration (int): The expiration time of the presigned URLs in seconds. Defaults to 3600.

        Returns:
            dict: A dictionary containing presigned URLs.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> my_image_document = lx.list_documents('my_image_collection', limit=1)[0]
            >>> lx.document.get_document_urls(document_id=my_image_document.document_id)
            {
                "object": "https://my-bucket.s3.amazonaws.com/path/to/object?...",
                "thumbnails": {
                    "256x256": "https://my-bucket.s3.amazonaws.com/path/to/thumbnail?..."
                }
            }

            >>> my_pdf_document = lx.list_documents('pdf_collection', limit=1)[0]
            >>> lx.document.get_document_urls(document_id=my_pdf_document.document_id)
            {
                "object": "https://my-bucket.s3.amazonaws.com/path/to/object?...",
            }
        """
        r = self.client.get(f"/documents/{document_id}/urls", params={"expiration": expiration})
        handle_response(r)
        return r.json()

    async def aget_document_urls(self, document_id: str, *, expiration: int = 3600) -> dict:
        """ Asynchronously get presigned URLs for a document.

        Args:
            document_id (str): The ID of the document to get presigned URL for.
            expiration (int): The expiration time of the presigned URLs in seconds. Defaults to 3600.

        Returns:
            dict: A dictionary containing presigned URLs.
        """
        r = await self.aclient.get(f"/documents/{document_id}/urls", params={"expiration": expiration})
        handle_response(r)
        return r.json()

    @staticmethod
    def _align_filenames(files, filenames: str | list[str] = None) -> tuple[list, list]:
        """ Align files and filenames. """
        # ensure files is a list
        if not isinstance(files, list):
            files = [files]

        # if filenames is a single string (or None), convert it to a list with repeated elements
        if isinstance(filenames, str) or filenames is None:
            filenames = [filenames] * len(files)

        if len(files) != len(filenames):
            raise ValueError("Length of filenames list must match length of files list")
        return files, filenames

    @staticmethod
    def _process_docs(docs: Document | dict | list[Document | dict]) -> list[dict]:
        """ Process documents into a list of json-serializable dictionaries. """
        processed_docs = []

        if isinstance(docs, (Document, dict)):
            docs = [docs]

        for doc in docs:
            if isinstance(doc, Document):
                processed_docs.append(doc.model_dump(mode='json'))
            elif isinstance(doc, dict):
                doc = Document(**doc)
                processed_docs.append(doc.model_dump(mode='json'))
            else:
                raise TypeError(f"Invalid type for doc item: {type(doc)} - must be Document or dict")

        return processed_docs

    @staticmethod
    def _process_files(files: str | Image.Image | list[str | Image.Image],
                       filenames: str | list[str] = None) -> list:
        processed_files = []
        files, filenames = DocumentClient._align_filenames(files, filenames)

        for file, filename in zip(files, filenames):
            if isinstance(file, str):
                with open(file, 'rb') as f:
                    file_content = f.read()
                if not filename:
                    filename = os.path.basename(file)
            elif isinstance(file, Image.Image):
                buffer = BytesIO()
                image_format = file.format or 'jpg'
                file.save(buffer, format=image_format)
                buffer.seek(0)
                file_content = buffer.getvalue()
                if not filename:
                    filename = f'image.{file.format.lower()}' if file.format else 'image.jpg'
            else:
                raise TypeError(f"Invalid type for file: {type(file)} - must be a str path to a file, or an "
                                f"`Image.Image` object")

            mime_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
            processed_files.append(('files', (filename, BytesIO(file_content), mime_type)))

        return processed_files



""" Client for interacting with the Indexes API. """

import io
import os
import mimetypes
from typing import Any, Optional, TYPE_CHECKING

import httpx
from PIL import Image

from lexy_py.exceptions import handle_response, LexyClientError
from .models import Index, IndexUpdate
from lexy_py.document.models import Document

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class IndexClient:
    """
    This class is used to interact with the Lexy Indexes API.

    Attributes:
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
    """

    def __init__(self, lexy_client: "LexyClient") -> None:
        self._lexy_client = lexy_client

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._lexy_client.aclient

    @property
    def client(self) -> httpx.Client:
        return self._lexy_client.client

    def list_indexes(self) -> list[Index]:
        """ Synchronously get a list of all indexes.

        Returns:
            list[Index]: A list of all indexes.
        """
        r = self.client.get("/indexes")
        handle_response(r)
        return [Index(**index, client=self._lexy_client) for index in r.json()]

    async def alist_indexes(self) -> list[Index]:
        """ Asynchronously get a list of all indexes.

        Returns:
            list[Index]: A list of all indexes.
        """
        r = await self.aclient.get("/indexes")
        handle_response(r)
        return [Index(**index, client=self._lexy_client) for index in r.json()]

    def get_index(self, index_id: str) -> Index:
        """ Synchronously get an index.

        Args:
            index_id (str): The ID of the index to get.

        Returns:
            Index: The index.
        """
        r = self.client.get(f"/indexes/{index_id}")
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    async def aget_index(self, index_id: str) -> Index:
        """ Asynchronously get an index.

        Args:
            index_id (str): The ID of the index to get.

        Returns:
            Index: The index.
        """
        r = await self.aclient.get(f"/indexes/{index_id}")
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    def add_index(self, index_id: str, description: Optional[str] = None,
                  index_table_schema: Optional[dict[str, Any]] = None,
                  index_fields: Optional[dict[str, Any]] = None) -> Index:
        """ Synchronously add an index.

        Args:
            index_id (str): The ID of the index to add.
            description (str, optional): A description of the index. Defaults to None.
            index_table_schema (dict[str, Any]): The schema of the index table. Defaults to None.
            index_fields (dict[str, Any]): The index fields that are created in the index table. These are typically
                the fields you want to populate using transformers. Defaults to None.

        Returns:
            Index: The index.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> code_index_fields = {
            ...     "code": {"type": "text"},
            ...     "code_embedding": {"type": "embedding", "extras": {"dims": 384, "model": "text.embeddings.minilm"}},
            ...     "n_lines": {"type": "int"},
            ... }
            >>> index = lx.index.add_index(index_id="code_index",
            ...                            description="Code embedding index",
            ...                            index_fields=code_index_fields)
        """
        if index_table_schema is None:
            index_table_schema = {}
        if index_fields is None:
            index_fields = {}
        data = {
            "index_id": index_id,
            "description": description,
            "index_table_schema": index_table_schema,
            "index_fields": index_fields
        }
        r = self.client.post("/indexes", json=data)
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    async def aadd_index(self, index_id: str, description: Optional[str] = None,
                         index_table_schema: Optional[dict[str, Any]] = None,
                         index_fields: Optional[dict[str, Any]] = None) -> Index:
        """ Asynchronously add an index.

        Args:
            index_id (str): The ID of the index to add.
            description (str, optional): A description of the index. Defaults to None.
            index_table_schema (dict[str, Any]): The schema of the index table. Defaults to None.
            index_fields (dict[str, Any]): The index fields that are created in the index table. These are typically
                the fields you want to populate using transformers. Defaults to None.

        Returns:
            Index: The index.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> code_index_fields = {
            ...     "code": {"type": "text"},
            ...     "code_embedding": {"type": "embedding", "extras": {"dims": 384, "model": "text.embeddings.minilm"}},
            ...     "n_lines": {"type": "int"},
            ... }
            >>> index = await lx.index.aadd_index(index_id="code_index",
            ...                                   description="Code embedding index",
            ...                                   index_fields=code_index_fields)
        """
        if index_table_schema is None:
            index_table_schema = {}
        if index_fields is None:
            index_fields = {}
        data = {
            "index_id": index_id,
            "description": description,
            "index_table_schema": index_table_schema,
            "index_fields": index_fields
        }
        r = await self.aclient.post("/indexes", json=data)
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    def delete_index(self, index_id: str, drop_table: bool = False) -> dict:
        """ Synchronously delete an index.

        Args:
            index_id (str): The ID of the index to delete.
            drop_table (bool, optional): Whether to drop the index table from the database. Defaults to False.
        """
        r = self.client.delete(f"/indexes/{index_id}", params={"drop_table": drop_table})
        handle_response(r)
        return r.json()

    async def adelete_index(self, index_id: str, drop_table: bool = False) -> dict:
        """ Asynchronously delete an index.

        Args:
            index_id (str): The ID of the index to delete.
            drop_table (bool, optional): Whether to drop the index table from the database. Defaults to False.
        """
        r = await self.aclient.delete(f"/indexes/{index_id}", params={"drop_table": drop_table})
        handle_response(r)
        return r.json()

    def update_index(self, index_id: str, description: Optional[str] = None,
                     index_table_schema: Optional[dict[str, Any]] = None,
                     index_fields: Optional[dict[str, Any]] = None) -> Index:
        """ Synchronously update an index.

        Args:
            index_id (str): The ID of the index to update.
            description (str, optional): The new description of the index.
            index_table_schema (dict[str, Any], optional): The new schema of the index table.
            index_fields (dict[str, Any], optional): The new value for index fields that are created in the index table.

        Returns:
            Index: The updated index.
        """
        index = IndexUpdate(
            description=description,
            index_table_schema=index_table_schema,
            index_fields=index_fields
        )
        r = self.client.patch(f"/indexes/{index_id}", json=index.model_dump(exclude_none=True))
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    async def aupdate_index(self, index_id: str, description: Optional[str] = None,
                            index_table_schema: Optional[dict[str, Any]] = None,
                            index_fields: Optional[dict[str, Any]] = None) -> Index:
        """ Asynchronously update an index.

        Args:
            index_id (str): The ID of the index to update.
            description (str, optional): The new description of the index.
            index_table_schema (dict[str, Any], optional): The new schema of the index table.
            index_fields (dict[str, Any], optional): The new value for index fields that are created in the index table.

        Returns:
            Index: The updated index.
        """
        index = IndexUpdate(
            description=description,
            index_table_schema=index_table_schema,
            index_fields=index_fields
        )
        r = await self.aclient.patch(f"/indexes/{index_id}", json=index.model_dump(exclude_none=True))
        handle_response(r)
        return Index(**r.json(), client=self._lexy_client)

    def list_index_records(self, index_id: str, document_id: Optional[str] = None) -> list[dict]:
        """ Synchronously get a list of all index records for an index.

        Args:
            index_id (str): The ID of the index to get records for.
            document_id (str, optional): The ID of a document to get records for.

        Returns:
            list[dict]: A list of all index records for an index.
        """
        params = {}
        if document_id:
            params["document_id"] = document_id
        r = self.client.get(f"/indexes/{index_id}/records", params=params)
        handle_response(r)
        return r.json()

    async def alist_index_records(self, index_id: str, document_id: Optional[str] = None) -> list[dict]:
        """ Asynchronously get a list of all index records for an index.

        Args:
            index_id (str): The ID of the index to get records for.
            document_id (str, optional): The ID of a document to get records for.

        Returns:
            list[dict]: A list of all index records for an index.
        """
        params = {}
        if document_id:
            params["document_id"] = document_id
        r = await self.aclient.get(f"/indexes/{index_id}/records", params=params)
        handle_response(r)
        return r.json()

    def query_index(self,
                    query_text: str = None,
                    query_image: Image.Image | str = None,
                    index_id: str = "default_text_embeddings",
                    query_field: str = "embedding",
                    k: int = 5,
                    return_fields: list[str] = None,
                    return_document: bool = False,
                    embedding_model: str = None) -> list[dict]:
        """ Synchronously query an index.

        Args:
            query_text (str): The query text.
            query_image (Image.Image | str): The query image. Can be a PIL Image object or a path to an image.
            index_id (str): The ID of the index to query. Defaults to "default_text_embeddings".
            query_field (str, optional): The field to query. Defaults to "embedding".
            k (int, optional): The number of records to return. Defaults to 5.
            return_fields (list[str], optional): The fields to return. Defaults to None, which returns all fields. To
                return fields from the linked document, use "document.<field_name>".
            return_document (bool, optional): Whether to return the document object. Defaults to False.
            embedding_model (str, optional): The name of the embedding model to use. Defaults to None, which uses the
                embedding model associated with `index_id.query_field`.

        Returns:
            list[dict]: The query results.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.query_index(query_text="Test Query")

            >>> lx.query_index(query_text="Test Query", return_fields=["my_index_field", "document.content"])

            >>> lx.query_index(query_image="test_image.jpg", index_id="my_image_index")

            >>> import httpx
            >>> from PIL import Image
            >>> img_url = 'https://getlexy.com/assets/images/dalle-agi.jpeg'
            >>> image = Image.open(httpx.get(img_url))
            >>> lx.query_index(query_image=image, index_id="my_image_index", k=3)
        """
        files, params = self._process_query_params(query_text=query_text,
                                                   query_image=query_image,
                                                   query_field=query_field,
                                                   k=k,
                                                   return_fields=return_fields,
                                                   return_document=return_document,
                                                   embedding_model=embedding_model)

        r = self.client.post(f"/indexes/{index_id}/records/query", files=files, params=params)
        handle_response(r)
        search_results = r.json()["search_results"]
        if return_document:
            for result in search_results:
                result["document"] = Document(**result["document"], client=self._lexy_client)
        return search_results

    async def aquery_index(self,
                           query_text: str = None,
                           query_image: Image.Image | str = None,
                           index_id: str = "default_text_embeddings",
                           query_field: str = "embedding",
                           k: int = 5,
                           return_fields: list[str] = None,
                           return_document: bool = False,
                           embedding_model: str = None) -> list[dict]:
        """ Asynchronously query an index.

        Args:
            query_text (str): The query text.
            query_image (Image.Image | str): The query image. Can be a PIL Image object or a path to an image.
            index_id (str): The ID of the index to query. Defaults to "default_text_embeddings".
            query_field (str, optional): The field to query. Defaults to "embedding".
            k (int, optional): The number of records to return. Defaults to 5.
            return_fields (list[str], optional): The fields to return. Defaults to None, which returns all fields. To
                return fields from the linked document, use "document.<field_name>".
            return_document (bool, optional): Whether to return the document object. Defaults to False.
            embedding_model (str, optional): The name of the embedding model to use. Defaults to None, which uses the
                embedding model associated with `index_id.query_field`.

        Returns:
            list[dict]: The query results.
        """
        files, params = self._process_query_params(query_text=query_text,
                                                   query_image=query_image,
                                                   query_field=query_field,
                                                   k=k,
                                                   return_fields=return_fields,
                                                   return_document=return_document,
                                                   embedding_model=embedding_model)

        r = await self.aclient.post(f"/indexes/{index_id}/records/query", files=files, params=params)
        handle_response(r)
        search_results = r.json()["search_results"]
        if return_document:
            for result in search_results:
                result["document"] = Document(**result["document"], client=self._lexy_client)
        return search_results

    @staticmethod
    def _process_query_params(query_text: str,
                              query_image: Image.Image | str,
                              query_field: str,
                              k: int,
                              return_fields: list[str],
                              return_document: bool,
                              embedding_model: str) -> tuple[dict, dict]:
        files = {}

        if query_text and query_image:
            raise LexyClientError("Please submit either 'query_text' or 'query_image', not both.")
        elif query_text:
            files["query_text"] = (None, query_text)
        elif query_image:
            if isinstance(query_image, str):
                image = Image.open(query_image)
                filename = os.path.basename(query_image)
            else:
                image = query_image
                filename = f'image.{image.format.lower()}' if image.format else 'image.jpg'
            buffer = io.BytesIO()
            image_format = image.format or "jpg"
            image.save(buffer, format=image_format)
            buffer.seek(0)
            mime_type = mimetypes.types_map.get(f".{image_format.lower()}", "application/octet-stream")
            files["query_image"] = (filename, buffer, mime_type)
        else:
            raise LexyClientError("Please submit either 'query_text' or 'query_image'.")

        if return_fields is None:
            return_fields = []
        params = {
            "query_field": query_field,
            "k": k,
            "return_fields": return_fields,
            "return_document": return_document
        }
        if embedding_model:
            params["embedding_model"] = embedding_model

        return files, params



from datetime import datetime, timedelta, timezone
from typing import Literal
from urllib.parse import parse_qs, urlparse


def presigned_url_is_expired(url: str, storage_service: str = 's3') -> bool:
    if storage_service == 's3':
        return signed_url_is_expired(url, svc='Amz')
    elif storage_service == 'gcs':
        return signed_url_is_expired(url, svc='Goog')
    elif storage_service == 'azure':
        raise NotImplementedError
    else:
        raise ValueError(f'Unsupported storage service: {storage_service} - '
                         f'must be one of the following: s3, gcs, azure')


def signed_url_is_expired(url: str, svc: Literal['Amz', 'Goog']):
    """Checks if a signed URL (V2 or V4) has expired.

    Args:
        url (str): The signed URL string.
        svc (Literal['Amz', 'Goog']): The service query parameter used to check for expiration in V4 signatures.
            Either 'Amz' for AWS or 'Goog' for GCP.

    Returns:
        True if the signed URL is expired, False otherwise.

    Raises:
        ValueError: If no expiration found in signed url.
    """
    query_params = parse_qs(urlparse(url).query)
    # V4 signature
    if f'X-{svc}-Expires' in query_params:
        expires_in = int(query_params[f'X-{svc}-Expires'][0])
        issued_at_dt = datetime.fromisoformat(query_params[f'X-{svc}-Date'][0])
        expires_at_dt = issued_at_dt + timedelta(seconds=expires_in)
        return expires_at_dt < datetime.now(tz=timezone.utc)
    # V2 signature
    elif 'Expires' in query_params:
        expires_at = int(query_params['Expires'][0])
        expires_at_dt = datetime.fromtimestamp(expires_at, tz=timezone.utc)
        return expires_at_dt < datetime.now(tz=timezone.utc)
    else:
        raise ValueError('No expiration found in signed url')



""" Client for interacting with the Collection API. """

from typing import Optional, TYPE_CHECKING

import httpx

from lexy_py.exceptions import handle_response
from .models import Collection, CollectionUpdate

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class CollectionClient:
    """
    This class is used to interact with the Lexy Collection API.

    Attributes:
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
    """

    def __init__(self, lexy_client: "LexyClient") -> None:
        self._lexy_client = lexy_client

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._lexy_client.aclient

    @property
    def client(self) -> httpx.Client:
        return self._lexy_client.client

    def list_collections(self) -> list[Collection]:
        """ Synchronously get a list of all collections.

        Returns:
            list[Collection]: A list of all collections.
        """
        r = self.client.get("/collections")
        handle_response(r)
        return [Collection(**collection, client=self._lexy_client) for collection in r.json()]

    async def alist_collections(self) -> list[Collection]:
        """ Asynchronously get a list of all collections.

        Returns:
            list[Collection]: A list of all collections.
        """
        r = await self.aclient.get("/collections")
        handle_response(r)
        return [Collection(**collection, client=self._lexy_client) for collection in r.json()]

    def get_collection(self,
                       *,
                       collection_name: str = None,
                       collection_id: str = None) -> Collection:
        """ Synchronously get a collection by name or ID.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to get.
            collection_id (str): The ID of the collection to get. If provided, `collection_name` will be ignored.

        Returns:
            Collection: The collection.

        Raises:
            ValueError: If neither `collection_name` nor `collection_id` are provided.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.get_collection(collection_id="70b8ce75")
            <Collection('test_collection', id='70b8ce75', description='My Test Collection')>

            >>> lx.get_collection(collection_name="test_collection")
            <Collection('test_collection', id='70b8ce75', description='My Test Collection')>
        """
        if collection_id:
            return self.get_collection_by_id(collection_id)
        elif collection_name:
            return self.get_collection_by_name(collection_name)
        else:
            raise ValueError("Either collection_id or collection_name must be provided.")

    async def aget_collection(self,
                              *,
                              collection_name: str = None,
                              collection_id: str = None) -> Collection:
        """ Asynchronously get a collection by name or ID.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to get.
            collection_id (str): The ID of the collection to get. If provided, `collection_name` will be ignored.

        Returns:
            Collection: The collection.

        Raises:
            ValueError: If neither `collection_name` nor `collection_id` are provided.
        """
        if collection_id:
            return await self.aget_collection_by_id(collection_id)
        elif collection_name:
            return await self.aget_collection_by_name(collection_name)
        else:
            raise ValueError("Either 'collection_name' or 'collection_id' must be provided.")

    def get_collection_by_id(self, collection_id: str) -> Collection:
        """ Synchronously get a collection by ID.

        Args:
            collection_id (str): The ID of the collection to get.

        Returns:
            Collection: The collection.
        """
        r = self.client.get(f"/collections/{collection_id}")
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    async def aget_collection_by_id(self, collection_id: str) -> Collection:
        """ Asynchronously get a collection by ID.

        Args:
            collection_id (str): The ID of the collection to get.

        Returns:
            Collection: The collection.
        """
        r = await self.aclient.get(f"/collections/{collection_id}")
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    def get_collection_by_name(self, collection_name: str) -> Collection:
        """ Synchronously get a collection by name.

        Args:
            collection_name (str): The name of the collection to get.

        Returns:
            Collection: The collection.
        """
        r = self.client.get(url="/collections", params={"collection_name": collection_name})
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    async def aget_collection_by_name(self, collection_name: str) -> Collection:
        """ Asynchronously get a collection by name.

        Args:
            collection_name (str): The name of the collection to get.

        Returns:
            Collection: The collection.
        """
        r = await self.aclient.get(url="/collections", params={"collection_name": collection_name})
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    def add_collection(self,
                       collection_name: str,
                       *,
                       description: Optional[str] = None,
                       config: Optional[dict] = None) -> Collection:
        """ Synchronously create a new collection.

        Args:
            collection_name (str): The name of the collection to create. Collection names must be unique.
            description (str, optional): The description of the collection. Defaults to None.
            config (dict, optional): The config of the collection. Defaults to None.

        Returns:
            Collection: The created collection.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.create_collection("test_collection", description="My Test Collection")
            <Collection('test_collection', id='70b8ce75', description='My Test Collection')>

            >>> lx.create_collection(collection_name="no_files",
            ...                      description="Collection without files",
            ...                      config={"store_files": False})
            <Collection('no_files', id='925d40ac', description='Collection without files')>
        """
        collection = Collection(
            collection_name=collection_name,
            description=description,
            config=config
        )
        r = self.client.post(url="/collections", json=collection.model_dump(exclude_none=True))
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    async def aadd_collection(self,
                              collection_name: str,
                              *,
                              description: Optional[str] = None,
                              config: Optional[dict] = None) -> Collection:
        """ Asynchronously create a new collection.

        Args:
            collection_name (str): The name of the collection to create. Collection names must be unique.
            description (str, optional): The description of the collection. Defaults to None.
            config (dict, optional): The config of the collection. Defaults to None.

        Returns:
            Collection: The created collection.
        """
        collection = Collection(
            collection_name=collection_name,
            description=description,
            config=config
        )
        r = await self.aclient.post(url="/collections", json=collection.model_dump(exclude_none=True))
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    def update_collection(self,
                          *,
                          collection_id: str,
                          collection_name: Optional[str] = None,
                          description: Optional[str] = None,
                          config: Optional[dict] = None) -> Collection:
        """ Synchronously update a collection.

        Args:
            collection_id (str): The ID of the collection to update.
            collection_name (str, optional): The updated name of the collection. Defaults to None.
            description (str, optional): The updated description of the collection. Defaults to None.
            config: (dict, optional): The updated config of the collection. Defaults to None.

        Returns:
            Collection: The updated collection.
        """
        collection = CollectionUpdate(
            collection_name=collection_name,
            description=description,
            config=config
        )
        r = self.client.patch(
            url=f"/collections/{collection_id}", json=collection.model_dump(exclude_none=True)
        )
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    async def aupdate_collection(self,
                                 *,
                                 collection_id: str,
                                 collection_name: Optional[str] = None,
                                 description: Optional[str] = None,
                                 config: Optional[dict] = None) -> Collection:
        """ Asynchronously update a collection.

        Args:
            collection_id (str): The ID of the collection to update.
            collection_name (str, optional): The updated name of the collection. Defaults to None.
            description (str, optional): The updated description of the collection. Defaults to None.
            config: (dict, optional): The updated config of the collection. Defaults to None.

        Returns:
            Collection: The updated collection.
        """
        collection = CollectionUpdate(
            collection_name=collection_name,
            description=description,
            config=config
        )
        r = await self.aclient.patch(
            url=f"/collections/{collection_id}", json=collection.model_dump(exclude_none=True)
        )
        handle_response(r)
        return Collection(**r.json(), client=self._lexy_client)

    def delete_collection(self,
                          *,
                          collection_name: str = None,
                          collection_id: str = None,
                          delete_documents: bool = False) -> dict:
        """ Synchronously delete a collection by name or ID.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to delete.
            collection_id (str): The ID of the collection to delete. If provided, `collection_name` will be ignored.
            delete_documents (bool, optional): Whether to delete the documents in the collection. Defaults to False.

        Raises:
            ValueError: If neither `collection_name` nor `collection_id` are provided.

        Examples:
            >>> from lexy_py import LexyClient
            >>> lx = LexyClient()
            >>> lx.delete_collection(collection_name="test_collection", delete_documents=True)
            {"msg": "Collection deleted", "collection_id": "70b8ce75", "documents_deleted": 3}
        """
        if collection_id:
            r = self.client.delete(
                url=f"/collections/{collection_id}", params={"delete_documents": delete_documents}
            )
        elif collection_name:
            r = self.client.delete(
                url="/collections", params={"collection_name": collection_name, "delete_documents": delete_documents}
            )
        else:
            raise ValueError("Either 'collection_name' or 'collection_id' must be provided.")
        handle_response(r)
        return r.json()

    async def adelete_collection(self,
                                 *,
                                 collection_name: str = None,
                                 collection_id: str = None,
                                 delete_documents: bool = False) -> dict:
        """ Asynchronously delete a collection by name or ID.

        If both `collection_name` and `collection_id` are provided, `collection_id` will be used.

        Args:
            collection_name (str): The name of the collection to delete.
            collection_id (str): The ID of the collection to delete. If provided, `collection_name` will be ignored.
            delete_documents (bool, optional): Whether to delete the documents in the collection. Defaults to False.

        Raises:
            ValueError: If neither `collection_name` nor `collection_id` are provided.
        """
        if collection_id:
            r = await self.aclient.delete(
                url=f"/collections/{collection_id}", params={"delete_documents": delete_documents}
            )
        elif collection_name:
            r = await self.aclient.delete(
                url="/collections", params={"collection_name": collection_name, "delete_documents": delete_documents}
            )
        else:
            raise ValueError("Either 'collection_name' or 'collection_id' must be provided.")
        handle_response(r)
        return r.json()



""" Lexy Python SDK """

from lexy_py.client import LexyClient
from lexy_py.filters import FilterBuilder

from lexy_py.binding.models import Binding
from lexy_py.collection.models import Collection
from lexy_py.document.models import Document
from lexy_py.index.models import Index
from lexy_py.transformer.models import Transformer



import json
from typing import Any


class FilterBuilder:
    """Helper class for creating filters.

    Attributes:
        conditions (list[dict]): A list of conditions to be combined. Each condition consists of a field, an operation,
            a value, and a negation option. `field` is a string representing the field to be matched. `operation` is a
            string representing the operation to be performed on the field. `value` is the value to be matched against.
            `negate` is a boolean representing whether the condition should be negated.

            Valid operations include the following.

            - `equals`
            - `equals_ci` (case-insensitive equals)
            - `less_than`
            - `less_than_or_equals`
            - `greater_than`
            - `greater_than_or_equals`
            - `contains`
            - `contains_ci` (case-insensitive contains)
            - `starts_with`
            - `starts_with_ci` (case-insensitive starts_with)
            - `ends_with`
            - `ends_with_ci` (case-insensitive ends_with)
            - `in`
        combination (str): The combination of conditions - either 'AND' or 'OR'

    Methods:
        include: Adds a condition to the filter object with a positive match
        exclude: Adds a condition to the filter object with a negative match
        to_dict: Returns the filter object as a dictionary
        to_json: Returns the filter object as a JSON string

    Examples:
        Restrict documents to those that contain the word 'mathematics' regardless of the case:

        >>> from lexy_py import FilterBuilder
        >>> builder = FilterBuilder()
        >>> builder.include("content", "contains_ci", "mathematics")
        >>> print(builder.to_json())
        {
            "conditions": [
                {"field": "content", "operation": "contains_ci", "value": "mathematics", "negate": false}
            ],
            "combination": "AND"
        }

        Restrict documents to those with a size less than 30,000 bytes and a file type that is not an image or video:

        >>> builder = FilterBuilder()
        >>> builder.include("meta.size", "less_than", 30000)
        >>> builder.exclude("meta.type", "in", ["image", "video"])
        >>> print(builder.to_json())
        {
            "conditions": [
                {"field": "meta.size", "operation": "less_than", "value": 30000, "negate": false},
                {"field": "meta.type", "operation": "in", "value": ["image", "video"], "negate": true}
            ],
            "combination": "AND"
        }

        Restrict documents to those where URL is not None and does not start with 'https://www.youtube.com':

        >>> builder = FilterBuilder()
        >>> builder.exclude("meta.url", "starts_with", "https://www.youtube.com")
        >>> builder.exclude("meta.url", "equals", None)
        >>> builder.to_dict()
        {
            "conditions": [
                {"field": "meta.url", "operation": "starts_with", "value": "https://www.youtube.com", "negate": True},
                {"field": "meta.url", "operation": "equals", "value": None, "negate": True}
            ],
            "combination": "AND"
        }
    """

    def __init__(self, conditions: list[dict] = None, combination: str = "AND"):
        self.conditions = conditions or []
        if combination.upper() not in ["AND", "OR"]:
            raise ValueError("Invalid combination - must be 'AND' or 'OR'")
        self.combination = combination.upper()

    def include(self, field: str, operation: str, value: Any):
        """Adds a condition to the filter object with a positive match

        Args:
            field: The field to be matched
            operation: The operation to be performed on the field. Must be one of the following.

                - `equals`
                - `equals_ci` (case-insensitive equals)
                - `less_than`
                - `less_than_or_equals`
                - `greater_than`
                - `greater_than_or_equals`
                - `contains`
                - `contains_ci` (case-insensitive contains)
                - `starts_with`
                - `starts_with_ci` (case-insensitive starts_with)
                - `ends_with`
                - `ends_with_ci` (case-insensitive ends_with)
                - `in`
            value: The value to be matched against
        """
        self.conditions.append({
            "field": field,
            "operation": operation,
            "value": value,
            "negate": False
        })
        return self

    def exclude(self, field: str, operation: str, value: Any):
        """Adds a condition to the filter object with a negative match

        Args:
            field: The field to be matched
            operation: The operation to be performed on the field. Must be one of the following.

                - `equals`
                - `equals_ci` (case-insensitive equals)
                - `less_than`
                - `less_than_or_equals`
                - `greater_than`
                - `greater_than_or_equals`
                - `contains`
                - `contains_ci` (case-insensitive contains)
                - `starts_with`
                - `starts_with_ci` (case-insensitive starts_with)
                - `ends_with`
                - `ends_with_ci` (case-insensitive ends_with)
                - `in`
            value: The value to be matched against
        """
        self.conditions.append({
            "field": field,
            "operation": operation,
            "value": value,
            "negate": True
        })
        return self

    def to_dict(self) -> dict:
        return {
            "conditions": self.conditions,
            "combination": self.combination
        }

    def to_json(self) -> str:
        return json.dumps({
            "conditions": self.conditions,
            "combination": self.combination
        })

    # TODO: implement from_dict and from_json methods to load an existing filter and modify it



from datetime import datetime
from enum import Enum
from typing import Any, Optional, TYPE_CHECKING

from pydantic import BaseModel, Field, PrivateAttr, field_validator, model_validator

from lexy_py.collection.models import CollectionModel, Collection
from lexy_py.index.models import IndexModel, Index
from lexy_py.transformer.models import TransformerModel, Transformer
from lexy_py.filters import FilterBuilder

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class BindingStatus(str, Enum):
    PENDING = "pending"
    ON = "on"
    OFF = "off"
    DETACHED = "detached"


# shared properties
class BindingBase(BaseModel):
    """ Binding base model """
    description: Optional[str] = None
    execution_params: Optional[dict[str, Any]] = Field(default={})
    transformer_params: Optional[dict[str, Any]] = Field(default={})
    filter: Optional[dict[str, Any]] = None
    status: Optional[BindingStatus] = Field(default=BindingStatus.PENDING)

    # if a FilterBuilder instance is provided for `filter`, convert it to a dict
    @field_validator('filter', mode='before')
    @classmethod
    def filter_to_dict(cls, value: Optional[FilterBuilder | dict]) -> Optional[dict]:
        if isinstance(value, FilterBuilder):
            return value.to_dict()
        return value


class BindingModel(BindingBase):
    """ Binding model """
    binding_id: int
    collection_id: str
    transformer_id: str
    index_id: str
    collection: CollectionModel
    transformer: TransformerModel
    index: IndexModel
    created_at: datetime
    updated_at: datetime

    def __repr__(self):
        return f"<Binding(" \
               f"id={self.binding_id}, " \
               f"status={self.status.name}, " \
               f"collection='{self.collection_id}', " \
               f"transformer='{self.transformer_id}', " \
               f"index='{self.index_id}')>"


class BindingCreate(BindingBase):
    """ Binding create model """
    collection_name: Optional[str] = None
    collection_id: Optional[str] = None
    transformer_name: Optional[str] = None
    transformer_id: Optional[str] = None
    index_name: Optional[str] = None
    index_id: Optional[str] = None

    # make sure either `_name` or `_id` is provided for collection, transformer, and index
    @model_validator(mode='before')
    def check_identifiers(cls, values):
        if not values.get('collection_name') and not values.get('collection_id'):
            raise ValueError('Either collection_name or collection_id must be provided')
        if not values.get('transformer_name') and not values.get('transformer_id'):
            raise ValueError('Either transformer_name or transformer_id must be provided')
        if not values.get('index_name') and not values.get('index_id'):
            raise ValueError('Either index_name or index_id must be provided')
        return values


class BindingUpdate(BindingBase):
    """ Binding update model """
    description: Optional[str] = None
    execution_params: Optional[dict[str, Any]] = None
    transformer_params: Optional[dict[str, Any]] = None
    filter: Optional[dict[str, Any]] = None
    status: Optional[BindingStatus] = None


class Binding(BindingModel):
    __doc__ = BindingModel.__doc__
    _client: Optional["LexyClient"] = PrivateAttr(default=None)

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._client = data.pop('client', None)
        self.collection: Collection = Collection(**self.collection.model_dump(), client=self._client)
        self.index: Index = Index(**self.index.model_dump(), client=self._client)
        self.transformer: Transformer = Transformer(**self.transformer.model_dump(), client=self._client)

    @property
    def client(self) -> "LexyClient":
        if not self._client:
            raise ValueError("API client has not been set.")
        return self._client



""" LexyClient exception classes. """

from typing import Optional

import httpx


class LexyClientError(Exception):
    """ Base exception class for LexyClient errors. """

    def __init__(self,
                 message: str,
                 response_data: Optional[dict] = None,
                 response: Optional[httpx.Response] = None):
        super().__init__(message)
        self.message = message
        self.response_data = response_data
        self.response = response

    def __str__(self):
        return f"{self.message}"


class LexyAPIError(LexyClientError):
    """ Exception class for API errors. """

    def __init__(self,
                 message: str = "Lexy API error",
                 response: Optional[httpx.Response] = None):
        if response:
            response_data = {
                "status_code": response.status_code,
                "text": response.text,
                "url": response.url,
            }
            message = f"{message} " \
                      f"'{response.status_code} {response.reason_phrase}' " \
                      f"for url '{response.url}'\n" \
                      f"\t{response.text}"
        else:
            response_data = None
        super().__init__(message, response_data, response)


class NotFoundError(LexyAPIError):
    """ Exception class for 404 errors. """

    def __init__(self,
                 message: str = "Lexy API error",
                 response: Optional[httpx.Response] = None):
        super().__init__(message, response)


def handle_response(response: httpx.Response) -> None:
    """ Handle API response. """
    if response.status_code == 404:
        raise NotFoundError(response=response)
    elif response.status_code >= 400:
        raise LexyAPIError(response=response)



""" Client for interacting with the Bindings API. """

from typing import Optional, TYPE_CHECKING

import httpx

from lexy_py.exceptions import handle_response
from lexy_py.filters import FilterBuilder
from .models import Binding, BindingCreate, BindingUpdate

if TYPE_CHECKING:
    from lexy_py.client import LexyClient


class BindingClient:
    """
    This class is used to interact with the Lexy Bindings API.

    Attributes:
        aclient (httpx.AsyncClient): Asynchronous API client.
        client (httpx.Client): Synchronous API client.
    """

    def __init__(self, lexy_client: "LexyClient") -> None:
        self._lexy_client = lexy_client

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._lexy_client.aclient

    @property
    def client(self) -> httpx.Client:
        return self._lexy_client.client

    def list_bindings(self) -> list[Binding]:
        """ Synchronously get a list of all bindings.

        Returns:
            list[Binding]: A list of all bindings.
        """
        r = self.client.get("/bindings")
        handle_response(r)
        return [Binding(**binding, client=self._lexy_client) for binding in r.json()]

    async def alist_bindings(self) -> list[Binding]:
        """ Asynchronously get a list of all bindings.

        Returns:
            list[Binding]: A list of all bindings.
        """
        r = await self.aclient.get("/bindings")
        handle_response(r)
        return [Binding(**binding, client=self._lexy_client) for binding in r.json()]

    def add_binding(self,
                    *,
                    collection_name: str = None,
                    collection_id: str = None,
                    transformer_name: str = None,
                    transformer_id: str = None,
                    index_name: str = None,
                    index_id: str = None,
                    description: Optional[str] = None,
                    execution_params: Optional[dict] = None,
                    transformer_params: Optional[dict] = None,
                    filters: Optional[dict | FilterBuilder] = None,
                    status: Optional[str] = None) -> Binding:
        """ Synchronously add a new binding.

        One of either `_name` or `_id` is required for each of `collection`, `transformer`, and `index`. If both
        `_name` and `_id` are provided, `_id` will be used.

        Args:
            collection_name (str): The name of the collection containing the source documents.
            collection_id (str): The ID of the collection containing the source documents.
            transformer_name (str): The name of the transformer that the binding will run.
            transformer_id (str): The ID of the transformer that the binding will run.
            index_name (str): The name of the index in which the binding will store its output.
            index_id (str): The ID of the index in which the binding will store its output.
            description (str, optional): A description of the binding.
            execution_params (dict, optional): Parameters to pass to the binding's execution function.
            transformer_params (dict, optional): Parameters to pass to the transformer.
            filters (dict | FilterBuilder, optional): Filters to apply to documents in the collection before running
                the transformer.
            status (str, optional): The status of the binding. Defaults to "pending".

        Returns:
            Binding: The created binding.

        Raises:
            ValueError: If neither `_name` nor `_id` is provided for each of `collection`, `transformer`, and `index`.

        Examples:
            Create a binding that runs the transformer with ID "image.embeddings.clip" on all image documents in
            the collection named "my_collection" and stores the output in the index with ID "image_embeddings".

            >>> from lexy_py import LexyClient, FilterBuilder
            >>> lx = LexyClient()
            >>> image_filter = FilterBuilder().include("meta.type", "equals", "image")
            >>> binding = lx.create_binding(
            ...     collection_name="my_collection",
            ...     transformer_id="image.embeddings.clip",
            ...     index_id="image_embeddings",
            ...     filters=image_filter
            ... )
        """
        # TODO: move execution_params and transformer_params logic to BindingCreate model
        if execution_params is None:
            execution_params = {}
        if transformer_params is None:
            transformer_params = {}
        binding = BindingCreate(
            collection_name=collection_name,
            collection_id=collection_id,
            transformer_name=transformer_name,
            transformer_id=transformer_id,
            index_name=index_name,
            index_id=index_id,
            description=description,
            execution_params=execution_params,
            transformer_params=transformer_params,
            filter=filters,
            status=status
        )
        r = self.client.post("/bindings", json=binding.model_dump(exclude_none=True))
        handle_response(r)
        return Binding(**r.json()["binding"], client=self._lexy_client)

    async def aadd_binding(self,
                           *,
                           collection_name: str = None,
                           collection_id: str = None,
                           transformer_name: str = None,
                           transformer_id: str = None,
                           index_name: str = None,
                           index_id: str = None,
                           description: Optional[str] = None,
                           execution_params: Optional[dict] = None,
                           transformer_params: Optional[dict] = None,
                           filters: Optional[dict | FilterBuilder] = None,
                           status: Optional[str] = None) -> Binding:
        """ Asynchronously add a new binding.

        One of either `_name` or `_id` is required for each of `collection`, `transformer`, and `index`. If both
        `_name` and `_id` are provided, `_id` will be used.

        Args:
            collection_name (str): The name of the collection containing the source documents.
            collection_id (str): The ID of the collection containing the source documents.
            transformer_name (str): The name of the transformer that the binding will run.
            transformer_id (str): The ID of the transformer that the binding will run.
            index_name (str): The name of the index in which the binding will store its output.
            index_id (str): The ID of the index in which the binding will store its output.
            description (str, optional): A description of the binding.
            execution_params (dict, optional): Parameters to pass to the binding's execution function.
            transformer_params (dict, optional): Parameters to pass to the transformer.
            filters (dict | FilterBuilder, optional): Filters to apply to documents in the collection before running
                the transformer.
            status (str, optional): The status of the binding. Defaults to "pending".

        Returns:
            Binding: The created binding.

        Raises:
            ValueError: If neither `_name` nor `_id` is provided for each of `collection`, `transformer`, and `index`.
        """
        # TODO: move execution_params and transformer_params logic to BindingCreate model
        if execution_params is None:
            execution_params = {}
        if transformer_params is None:
            transformer_params = {}
        binding = BindingCreate(
            collection_name=collection_name,
            collection_id=collection_id,
            transformer_name=transformer_name,
            transformer_id=transformer_id,
            index_name=index_name,
            index_id=index_id,
            description=description,
            execution_params=execution_params,
            transformer_params=transformer_params,
            filter=filters,
            status=status
        )
        r = await self.aclient.post("/bindings", json=binding.model_dump(exclude_none=True))
        handle_response(r)
        return Binding(**r.json()["binding"], client=self._lexy_client)

    def get_binding(self, binding_id: int) -> Binding:
        """ Synchronously get a binding.

        Args:
            binding_id (int): The ID of the binding to get.

        Returns:
            Binding: The binding.
        """
        r = self.client.get(f"/bindings/{binding_id}")
        handle_response(r)
        return Binding(**r.json(), client=self._lexy_client)

    async def aget_binding(self, binding_id: int) -> Binding:
        """ Asynchronously get a binding.

        Args:
            binding_id (int): The ID of the binding to get.

        Returns:
            Binding: The binding.
        """
        r = await self.aclient.get(f"/bindings/{binding_id}")
        handle_response(r)
        return Binding(**r.json(), client=self._lexy_client)

    def update_binding(self,
                       binding_id: int,
                       *,
                       description: Optional[str] = None,
                       execution_params: Optional[dict] = None,
                       transformer_params: Optional[dict] = None,
                       filters: Optional[dict | FilterBuilder] = None,
                       status: Optional[str] = None) -> Binding:
        """ Synchronously update a binding.

        Args:
            binding_id (int): The ID of the binding to update.
            description (str, optional): A description of the binding.
            execution_params (dict, optional): Parameters to pass to the binding's execution function.
            transformer_params (dict, optional): Parameters to pass to the transformer.
            filters (dict | FilterBuilder, optional): Filters to apply to documents in the collection before running
                the transformer. Set to an empty dict to remove any existing filter.
            status (str, optional): The status of the binding.

        Returns:
            Binding: The updated binding.
        """
        binding = BindingUpdate(
            description=description,
            execution_params=execution_params,
            transformer_params=transformer_params,
            filter=filters,
            status=status
        )
        json_payload = binding.model_dump(exclude_none=True)
        if json_payload["filter"] == {}:  # explicitly removed filters
            json_payload["filter"] = None
        r = self.client.patch(f"/bindings/{binding_id}", json=json_payload)
        handle_response(r)
        return Binding(**r.json()["binding"], client=self._lexy_client)

    async def aupdate_binding(self,
                              *,
                              binding_id: int,
                              description: Optional[str] = None,
                              execution_params: Optional[dict] = None,
                              transformer_params: Optional[dict] = None,
                              filters: Optional[dict | FilterBuilder] = None,
                              status: Optional[str] = None) -> Binding:
        """ Asynchronously update a binding.

        Args:
            binding_id (int): The ID of the binding to update.
            description (str, optional): A description of the binding.
            execution_params (dict, optional): Parameters to pass to the binding's execution function.
            transformer_params (dict, optional): Parameters to pass to the transformer.
            filters (dict | FilterBuilder, optional): Filters to apply to documents in the collection before running
                the transformer. Set to an empty dict to remove any existing filter.
            status (str, optional): The status of the binding.

        Returns:
            Binding: The updated binding.
        """
        binding = BindingUpdate(
            description=description,
            execution_params=execution_params,
            transformer_params=transformer_params,
            filter=filters,
            status=status
        )
        json_payload = binding.model_dump(exclude_none=True)
        if json_payload["filter"] == {}:  # explicitly removed filters
            json_payload["filter"] = None
        r = await self.aclient.patch(f"/bindings/{binding_id}", json=json_payload)
        handle_response(r)
        return Binding(**r.json()["binding"], client=self._lexy_client)

    def delete_binding(self, binding_id: int) -> dict:
        """ Synchronously delete a binding.

        Args:
            binding_id (int): The ID of the binding to delete.
        """
        r = self.client.delete(f"/bindings/{binding_id}")
        handle_response(r)
        return r.json()

    async def adelete_binding(self, binding_id: int) -> dict:
        """ Asynchronously delete a binding.

        Args:
            binding_id (int): The ID of the binding to delete.
        """
        r = await self.aclient.delete(f"/bindings/{binding_id}")
        handle_response(r)
        return r.json()
